//******************************************************************************************
//* Apollonian Gasket, by Paul Nylander, bugman123.com, 7/3/06
//******************************************************************************************

camera{orthographic location <0,0,-1>}

//Basic Functions
#macro square(X) X*X #end

//Complex Algebra
#macro Complex(X,Y) <X,Y> #end
#declare I=Complex(0,1);
#macro Re(Z) Z.x #end
#macro Im(Z) Z.y #end
#macro Conjugate(Z) Complex(Re(Z),-Im(Z)) #end
#macro Abs(Z) vlength(Z) #end // sqrt(square(Re(Z))+square(Im(Z)))
#macro Arg(Z) atan2(Im(Z),Re(Z)) #end
#macro Pow(Z,n) #local r=Abs(Z); #if(r=0) #local z2=Complex(0,0); #else #local theta=n*Arg(Z); #local z2=pow(r,n)*Complex(cos(theta),sin(theta)); #end z2 #end
#macro Exp(Z) exp(Re(Z))*Complex(cos(Im(Z)),sin(Im(Z))) #end
#macro Mult(z1,z2) Complex(Re(z1)*Re(z2)-Im(z1)*Im(z2),Im(z1)*Re(z2)+Re(z1)*Im(z2)) #end
#macro Div(z1,z2) Mult(z1,Pow(z2,-1)) #end

//Chain Functions
#macro Chain(Z,r) array[2]{Z,<r,0>} #end
#macro zchain(c) c[0] #end
#macro rchain(c) c[1].x #end
#macro nth(i,chains) Chain(chains[i][0],chains[i][1].x) #end // kind of ugly
#macro setnth(i,chains,c) #declare chains[i][0]=c[0]; #declare chains[i][1]=c[1]; chains #end

//Initial Chains: simply modify these initial chains to get different Schottky circles
#declare chains0=array[4][2]; #declare i=0;
#while(i<3) #declare theta=pi*(1/2+2*i)/3; #declare chains0=setnth(i,chains0,Chain(2*Exp(I*theta),sqrt(3))); #declare i=i+1; #end
#declare chains0=setnth(3,chains0,Chain(Complex(0,0),2-sqrt(3)));

//Calculations
#macro Reflect(c2,c1) #local z1=zchain(c1); #local z2=zchain(c2); #local r1=rchain(c1); #local r2=rchain(c2);
 #local a=Re(Div(Complex(square(r1),0),Mult(z2-z1,Conjugate(z2-z1))-Complex(square(r2),0)));
 Chain(z1+a*(z2-z1),a*r2)
#end
#macro children(chains1) #local n=dimension_size(chains1,1); #local chains2=array[n*(n-1)][2]; #local i=0; #local k=0;
 #while(i<n) #local j=0;
  #while(j<n)
   #if(j!=i) #local chains2=setnth(k,chains2,Reflect(nth(j,chains1),nth(i,chains1))); #local k=k+1; #end
   #local j=j+1;
  #end
  #local i=i+1;
 #end
 chains2
#end
#declare chains1=children(chains0);
#declare chains2=children(chains1);

//Draw Circles
#macro drawcircles(depth,chains) #local n=dimension_size(chains,1); #local i=0; #while(i<n)
 #local Z=chains[i][0]; #local r=chains[i][1].x;
 disc{<Re(Z),Im(Z),-1e-6*depth>,<0,0,1>,r
  pigment{function{(depth+1)/3} color_map{[0 rgb <0,0,0>][1/3 rgb <0,0,1>][2/3 rgb <1,0,1>][1 rgb <1,1,1>]}}
  finish{ambient 1}
 }
 #local i=i+1;
#end #end
drawcircles(0,chains0)
drawcircles(1,chains1)
drawcircles(2,chains2)
